# Copyright (c) 2023 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import client as http_client
import json

from oslo_log import log as logging
import requests

from manila import exception

LOG = logging.getLogger(__name__)


class StorageObjectManager(object):

    def __init__(self,
                 host_url,
                 username,
                 password,
                 export_path,
                 certificate_path=None,
                 verify_ssl_cert=False):
        self.host_url = host_url
        self.base_url = host_url + '/rest'
        self.rest_username = username
        self.rest_password = password
        self.rest_token = None
        self.got_token = False
        self.export_path = export_path
        self.verify_certificate = verify_ssl_cert
        self.certificate_path = certificate_path

    def _get_headers(self):
        if self.got_token:
            return {"Content-type": "application/json",
                    "Accept": "application/json",
                    "Authorization": "Bearer " + self.rest_token}
        else:
            return {"Content-type": "application/json",
                    "Accept": "application/json"}

    def execute_powerflex_get_request(self, url, **url_params):
        request = url % url_params
        res = requests.get(request,
                           headers=self._get_headers(),
                           verify=self._get_verify_cert())
        res = self._check_response(res, request, "GET")
        response = res.json()

        return res, response

    def execute_powerflex_post_request(self, url, params=None, **url_params):
        if not params:
            params = {}
        request = url % url_params
        res = requests.post(request,
                            data=json.dumps(params),
                            headers=self._get_headers(),
                            verify=self._get_verify_cert())
        res = self._check_response(res, request, "POST", params)
        response = None
        try:
            response = res.json()
        except ValueError:
            # Particular case for get_storage_pool_id which is not
            # a json object but a string
            response = res
        return res, response

    def execute_powerflex_delete_request(self, url, **url_params):
        request = url % url_params
        res = requests.delete(request,
                              headers=self._get_headers(),
                              verify=self._get_verify_cert())
        res = self._check_response(res, request, "DELETE")
        return res

    def execute_powerflex_patch_request(self, url, params=None, **url_params):
        if not params:
            params = {}
        request = url % url_params
        res = requests.patch(request,
                             data=json.dumps(params),
                             headers=self._get_headers(),
                             verify=self._get_verify_cert())
        res = self._check_response(res, request, "PATCH")
        return res

    def _check_response(self,
                        response,
                        request,
                        request_type,
                        params=None):
        login_url = "/auth/login"

        if (response.status_code == http_client.UNAUTHORIZED or
                response.status_code == http_client.FORBIDDEN):
            LOG.info("Dell PowerFlex token is invalid, going to re-login "
                     "and get a new one.")
            login_request = self.base_url + login_url
            verify_cert = self._get_verify_cert()
            self.got_token = False
            payload = json.dumps({"username": self.rest_username,
                                  "password": self.rest_password})
            res = requests.post(login_request,
                                headers=self._get_headers(),
                                data=payload,
                                verify=verify_cert)
            if (res.status_code == http_client.UNAUTHORIZED or
                    res.status_code == http_client.FORBIDDEN):
                message = ("PowerFlex REST API access is still forbidden or "
                           "unauthorized, there might be an issue with your "
                           "credentials.")
                LOG.error(message)
                raise exception.NotAuthorized()
            else:
                token = res.json()["access_token"]
                self.rest_token = token
                self.got_token = True
                LOG.info("Going to perform request again %s with valid token.",
                         request)
                if (request_type == "GET"):
                    response = requests.get(request,
                                            headers=self._get_headers(),
                                            verify=verify_cert)
                elif (request_type == "POST"):
                    response = requests.post(request,
                                             headers=self._get_headers(),
                                             data=json.dumps(params),
                                             verify=verify_cert)
                elif (request_type == "DELETE"):
                    response = requests.delete(request,
                                               headers=self._get_headers(),
                                               verify=verify_cert)
                elif (request_type == "PATCH"):
                    response = requests.patch(request,
                                              headers=self._get_headers(),
                                              data=json.dumps(params),
                                              verify=verify_cert)
                level = logging.DEBUG
                if response.status_code != http_client.OK:
                    level = logging.ERROR
                LOG.log(level,
                        "REST REQUEST: %s with params %s",
                        request,
                        json.dumps(params))
                LOG.log(level,
                        "REST RESPONSE: %s with params %s",
                        response.status_code,
                        response.text)
        return response

    def _get_verify_cert(self):
        verify_cert = False
        if self.verify_certificate:
            verify_cert = self.certificate_path
        return verify_cert

    def create_filesystem(self, storage_pool_id, nas_server, name, size):
        """Creates a filesystem.

        :param nas_server: name of the nas_server
        :param name: name of the filesystem
        :param size: size in GiB
        :return: ID of the filesystem if created successfully
        """
        nas_server_id = self.get_nas_server_id(nas_server)
        params = {
            "name": name,
            "size_total": size,
            "storage_pool_id": storage_pool_id,
            "nas_server_id": nas_server_id
            }
        url = f'{self.base_url}/v1/file-systems'
        res, response = self.execute_powerflex_post_request(url, params)
        if res.status_code == 201:
            return response["id"]

    def create_nfs_export(self, filesystem_id, name):
        """Creates an NFS export.

        :param filesystem_id: ID of the filesystem on which
                              the export will be created
        :param name: name of the NFS export
        :return: ID of the export if created successfully
        """
        params = {
            "file_system_id": filesystem_id,
            "path": "/" + str(name),
            "name": name
            }
        url = f'{self.base_url}/v1/nfs-exports'
        res, response = self.execute_powerflex_post_request(url, params)
        if res.status_code == 201:
            return response["id"]

    def delete_filesystem(self, filesystem_id):
        """Deletes a filesystem and all associated export.

        :param filesystem_id: ID of the filesystem to delete
        :return: True if deleted successfully
        """
        url = f'{self.base_url}/v1/file-systems/{filesystem_id}'
        res = self.execute_powerflex_delete_request(url)
        return res.status_code == 204

    def create_snapshot(self, name, filesystem_id):
        """Creates a snapshot of a filesystem.

        :param name: name of the snapshot
        :param filesystem_id: ID of the filesystem
        :return: ID of the snapshot if created successfully
        """
        params = {
            "name": name
            }
        url = f'{self.base_url}/v1/file-systems/{filesystem_id}/snapshot'
        res, response = self.execute_powerflex_post_request(url, params)
        return res.status_code == 201

    def get_nas_server_id(self, nas_server):
        """Retrieves the NAS server ID.

        :param nas_server: NAS server name
        :return: ID of the NAS server if success
        """
        url = f'{self.base_url}/v1/nas-servers?select=id&name=eq.{nas_server}'
        res, response = self.execute_powerflex_get_request(url)
        if res.status_code == 200:
            return response[0]['id']

    def get_nfs_export_name(self, export_id):
        """Retrieves NFS Export name.

        :param export_id: ID of the NFS export
        :return: path of the NFS export if success
        """
        url = f'{self.base_url}/v1/nfs-exports/{export_id}?select=*'
        res, response = self.execute_powerflex_get_request(url)
        if res.status_code == 200:
            return response["name"]

    def get_filesystem_id(self, name):
        """Retrieves an ID for a filesystem.

        :param name: name of the filesystem
        :return: ID of the filesystem if success
        """
        url = f'{self.base_url}/v1/file-systems?select=id&name=eq.{name}'
        res, response = self.execute_powerflex_get_request(url)
        if res.status_code == 200:
            return response[0]['id']

    def get_nfs_export_id(self, name):
        """Retrieves NFS Export ID.

        :param name: name of the NFS export
        :return: id of the NFS export if success
        """
        url = f'{self.base_url}/v1/nfs-exports?select=id&name=eq.{name}'
        res, response = self.execute_powerflex_get_request(url)
        if res.status_code == 200:
            return response[0]['id']

    def get_storage_pool_id(self, protection_domain, storage_pool):
        """Retrieves the Storage Pool ID.

        :param protection_domain: protection domain name
        :param storage_pool: storage pool name
        :return: ID of the storage pool if success
        """
        params = {
            "protectionDomainName": protection_domain,
            "name": storage_pool
            }
        url = (f'{self.host_url}/api/types/StoragePool/instances/'
               'action/queryIdByKey')
        res, response = self.execute_powerflex_post_request(url, params)
        if res.status_code == 200:
            return response

    def set_export_access(self, export_id, rw_hosts, ro_hosts):
        """Sets the authorization access on the export.

        :param export_id: NFS export ID
        :param rw_hosts: a set of RW hosts
        :param ro_hosts: a set of RO hosts
        :return: True if operation succeeded
        """
        params = {
            "read_only_hosts": list(ro_hosts),
            "read_write_root_hosts": list(rw_hosts)
            }
        url = f'{self.base_url}/v1/nfs-exports/{export_id}'
        res = self.execute_powerflex_patch_request(url, params)
        return res.status_code == 204

    def extend_export(self, export_id, new_size):
        """Extends the size of a share to a new size.

        :param export_id: ID of the NFS export
        :param new_size: new size to allocate in bytes
        :return: True if extended successfully
        """
        params = {
            "size_total": new_size
            }
        url = f'{self.base_url}/v1/file-systems/{export_id}'
        res = self.execute_powerflex_patch_request(url, params)
        return res.status_code == 204

    def get_fsid_from_export_name(self, name):
        """Retieves the Filesystem ID used by an export.

        :param name: name of the export
        :return: ID of the Filesystem which owns the export
        """
        url = (f'{self.base_url}/v1/nfs-exports'
               f'?select=file_system_id&name=eq.{name}')
        res, response = self.execute_powerflex_get_request(url)
        if res.status_code == 200:
            return response[0]['file_system_id']

    def get_fsid_from_snapshot_name(self, snapshot_name):
        """Retrieves the Filesystem ID used by a snapshot.

        :param snapshot_name: Name of the snapshot
        :return: ID of the parent filesystem of the snapshot
        """
        url = (f'{self.base_url}/v1/file-systems'
               f'?select=id&name=eq.{snapshot_name}')
        res, response = self.execute_powerflex_get_request(url)
        if res.status_code == 200:
            return response[0]['id']

    def get_storage_pool_spare_percentage(self, storage_pool_id):
        """Retrieves the spare capacity percentage of the storage pool.

        :param storage_pool_id: ID of the storage pool
        :return: Spare capacity percentage of the storage pool
        """
        url = f'{self.host_url}/api/instances/StoragePool::{storage_pool_id}'
        res, response = self.execute_powerflex_get_request(url)
        if res.status_code == 200:
            return response['sparePercentage']

    def get_storage_pool_statistic(self, storage_pool_id):
        """Retrieves the spare capacity percentage of the storage pool.

        :param storage_pool_id: ID of the storage pool
        :return: Statistics of the storage pool
        """
        url = (f'{self.host_url}/api/instances/StoragePool::{storage_pool_id}'
               '/relationships/Statistics')
        res, response = self.execute_powerflex_get_request(url)
        if res.status_code == 200:
            statistics = {
                "maxCapacityInKb": response['maxCapacityInKb'],
                "capacityInUseInKb": response['capacityInUseInKb'],
                "netUnusedCapacityInKb": response['netUnusedCapacityInKb'],
                "primaryVacInKb": response['primaryVacInKb'],
            }
        return statistics

    def get_nas_server_interfaces(self, nas_server_id):
        """Retrieves the file interfaces for a given na_server.

        :param nas_server_id: ID of the NAS server
        :return: file interfaces of the NAS server
        """
        url = (f'{self.base_url}/v1/file-interfaces'
               f'?select=ip_address&nas_server_id=eq.{nas_server_id}')
        res, response = self.execute_powerflex_get_request(url)
        if res.status_code == 200:
            return [i['ip_address'] for i in response]



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy

from manila.policies import base


BASE_POLICY_NAME = 'share_backup:%s'

DEPRECATED_REASON = """
The share backup API now supports system scope and default roles.
"""

deprecated_backup_create = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'create',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since='2023.2/Bobcat'
)
deprecated_backup_get = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'get',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since='2023.2/Bobcat',
)
deprecated_backup_get_all = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'get_all',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since='2023.2/Bobcat',
)
deprecated_get_all_project = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'get_all_project',
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since='2023.2/Bobcat',
)
deprecated_backup_restore = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'restore',
    check_str=base.RULE_ADMIN_OR_OWNER,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since='2023.2/Bobcat',
)
deprecated_backup_update = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'update',
    check_str=base.RULE_ADMIN_OR_OWNER,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since='2023.2/Bobcat',
)
deprecated_backup_delete = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'delete',
    check_str=base.RULE_ADMIN_OR_OWNER,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since='2023.2/Bobcat',
)
deprecated_backup_reset_status = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'reset_status',
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since='2023.2/Bobcat',
)


share_backup_policies = [
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'create',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Create share backup.",
        operations=[
            {
                'method': 'POST',
                'path': '/share-backups'
            }
        ],
        deprecated_rule=deprecated_backup_create,
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'get',
        check_str=base.ADMIN_OR_PROJECT_READER,
        scope_types=['project'],
        description="Get share backup.",
        operations=[
            {
                'method': 'GET',
                'path': '/share-backups/{backup_id}'
            }
        ],
        deprecated_rule=deprecated_backup_get,
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'get_all',
        check_str=base.ADMIN_OR_PROJECT_READER,
        scope_types=['project'],
        description="Get all share backups.",
        operations=[
            {
                'method': 'GET',
                'path': '/share-backups'
            },
            {
                'method': 'GET',
                'path': '/share-backups/detail'
            },
            {
                'method': 'GET',
                'path': '/share-backups/detail?share_id=(share_id}',
            },
        ],
        deprecated_rule=deprecated_backup_get_all,
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'get_all_project',
        check_str=base.ADMIN,
        scope_types=['project'],
        description="Get share backups of all projects.",
        operations=[
            {
                'method': 'GET',
                'path': '/share-backups?all_tenants=1'
            },
            {
                'method': 'GET',
                'path': '/share-backups/detail?all_tenants=1'
            }
        ],
        deprecated_rule=deprecated_get_all_project
    ),

    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'restore',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Restore a share backup.",
        operations=[
            {
                'method': 'POST',
                'path': '/share-backups/{backup_id}/action'
            }
        ],
        deprecated_rule=deprecated_backup_restore,
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'reset_status',
        check_str=base.ADMIN,
        scope_types=['project'],
        description="Reset status.",
        operations=[
            {
                'method': 'POST',
                'path': '/share-backups/{backup_id}/action',
            }
        ],
        deprecated_rule=deprecated_backup_reset_status
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'update',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Update a share backup.",
        operations=[
            {
                'method': 'PUT',
                'path': '/share-backups/{backup_id}',
            }
        ],
        deprecated_rule=deprecated_backup_update,
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'delete',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Force Delete a share backup.",
        operations=[
            {
                'method': 'DELETE',
                'path': '/share-backups/{backup_id}'
            }
        ],
        deprecated_rule=deprecated_backup_delete,
    ),

]


def list_rules():
    return share_backup_policies



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add_transfers

Revision ID: 1e2d600bf972
Revises: c476aeb186ec
Create Date: 2022-05-30 16:37:18.325464

"""

# revision identifiers, used by Alembic.
revision = '1e2d600bf972'
down_revision = 'c476aeb186ec'

from alembic import op
from oslo_log import log
import sqlalchemy as sa

LOG = log.getLogger(__name__)


def upgrade():
    context = op.get_context()
    mysql_dl = context.bind.dialect.name == 'mysql'
    datetime_type = (sa.dialects.mysql.DATETIME(fsp=6)
                     if mysql_dl else sa.DateTime)

    try:
        op.create_table(
            'transfers',
            sa.Column('id', sa.String(36), primary_key=True, nullable=False),
            sa.Column('created_at', datetime_type),
            sa.Column('updated_at', datetime_type),
            sa.Column('deleted_at', datetime_type),
            sa.Column('deleted', sa.String(36), default='False'),
            sa.Column('resource_id', sa.String(36), nullable=False),
            sa.Column('resource_type', sa.String(255), nullable=False),
            sa.Column('display_name', sa.String(255)),
            sa.Column('salt', sa.String(255)),
            sa.Column('crypt_hash', sa.String(255)),
            sa.Column('expires_at', datetime_type),
            sa.Column('source_project_id', sa.String(255), nullable=True),
            sa.Column('destination_project_id', sa.String(255), nullable=True),
            sa.Column('accepted', sa.Boolean, default=False),
            mysql_engine='InnoDB',
            mysql_charset='utf8',
        )
    except Exception:
        LOG.error("Table |%s| not created!", 'transfers')
        raise


def downgrade():
    try:
        op.drop_table('transfers')
    except Exception:
        LOG.error("transfers table not dropped")
        raise



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Ssh utilities."""

import logging
import os

from eventlet import pools
from oslo_config import cfg
from oslo_log import log
from oslo_utils.secretutils import md5

from manila import exception
from manila.i18n import _


try:
    import paramiko
except ImportError:
    paramiko = None

CONF = cfg.CONF
LOG = log.getLogger(__name__)
if getattr(CONF, 'debug', False):
    logging.getLogger("paramiko").setLevel(logging.DEBUG)


def get_fingerprint(self):
    """Patch paramiko

    This method needs to be patched to allow paramiko to work under FIPS.
    Until the patch to do this merges, patch paramiko here.

    TODO(carloss) Remove this when paramiko is patched.
    See https://github.com/paramiko/paramiko/pull/1928
    """
    return md5(self.asbytes(), usedforsecurity=False).digest()


if paramiko is None:
    raise exception.RequirementMissing(req='paramiko')

paramiko.pkey.PKey.get_fingerprint = get_fingerprint


class SSHPool(pools.Pool):
    """A simple eventlet pool to hold ssh connections."""

    def __init__(self, ip, port, conn_timeout, login, password=None,
                 privatekey=None, *args, **kwargs):
        self.ip = ip
        self.port = port
        self.login = login
        self.password = password
        self.conn_timeout = conn_timeout if conn_timeout else None
        self.path_to_private_key = privatekey
        super(SSHPool, self).__init__(*args, **kwargs)

    def create(self):  # pylint: disable=method-hidden
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        look_for_keys = True
        if self.path_to_private_key:
            self.path_to_private_key = os.path.expanduser(
                self.path_to_private_key)
            look_for_keys = False
        elif self.password:
            look_for_keys = False
        try:
            LOG.debug("ssh.connect: ip: %s, port: %s, look_for_keys: %s, "
                      "timeout: %s, banner_timeout: %s",
                      self.ip,
                      self.port,
                      look_for_keys,
                      self.conn_timeout,
                      self.conn_timeout)
            ssh.connect(self.ip,
                        port=self.port,
                        username=self.login,
                        password=self.password,
                        key_filename=self.path_to_private_key,
                        look_for_keys=look_for_keys,
                        timeout=self.conn_timeout,
                        banner_timeout=self.conn_timeout)
            if self.conn_timeout:
                transport = ssh.get_transport()
                transport.set_keepalive(self.conn_timeout)
            return ssh
        except Exception as e:
            msg = _("Check whether private key or password are correctly "
                    "set. Error connecting via ssh: %s") % e
            LOG.error(msg)
            raise exception.SSHException(msg)

    def get(self):
        """Return an item from the pool, when one is available.

        This may cause the calling greenthread to block. Check if a
        connection is active before returning it. For dead connections
        create and return a new connection.
        """
        if self.free_items:
            conn = self.free_items.popleft()
            if conn:
                if conn.get_transport().is_active():
                    return conn
                else:
                    conn.close()
            return self.create()
        if self.current_size < self.max_size:
            created = self.create()
            self.current_size += 1
            return created
        return self.channel.get()

    def remove(self, ssh):
        """Close an ssh client and remove it from free_items."""
        ssh.close()
        if ssh in self.free_items:
            self.free_items.remove(ssh)
            if self.current_size > 0:
                self.current_size -= 1



# Copyright 2023 NetApp, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
import requests
from requests.adapters import HTTPAdapter
from requests import auth
from urllib3.util import retry

from manila import exception
from manila.scheduler.weighers import base_host

ACTIVE_IQ_WEIGHER_GROUP = 'netapp_active_iq'


active_iq_weight_opts = [
    cfg.HostAddressOpt('aiq_hostname',
                       help='The hostname (or IP address) for the Active IQ.'),
    cfg.PortOpt('aiq_port',
                help=('The TCP port to use for communication with the Active '
                      'IQ. If not specified, the weigher driver will use 80 '
                      'for HTTP and 443 for HTTPS.')),
    cfg.StrOpt('aiq_transport_type',
               default='https',
               choices=['http', 'https'],
               help=('The transport protocol used when communicating with '
                     'the Active IQ. Valid values are '
                     'http or https.')),
    cfg.BoolOpt('aiq_ssl_verify',
                default=False,
                help='Verifying the SSL certificate. Default is False.'),
    cfg.StrOpt('aiq_ssl_cert_path',
               help=("The path to a CA_BUNDLE file or directory with "
                     "certificates of trusted CA. If set to a directory, it "
                     "must have been processed using the c_rehash utility "
                     "supplied with OpenSSL. If not informed, it will use the "
                     "Mozilla's carefully curated collection of Root "
                     "Certificates for validating the trustworthiness of SSL "
                     "certificates.")),
    cfg.StrOpt('aiq_username',
               help=('Administrative user account name used to access the '
                     'Active IQ.')),
    cfg.StrOpt('aiq_password',
               help=('Password for the administrative user account '
                     'specified in the aiq_username option.'),
               secret=True),
    cfg.IntOpt('aiq_eval_method',
               default=0,
               help='Integer indicator of which evaluation method, defaults '
                    'to 0 (0 - by index, 1 - normalized value, 2 - by '
                    'literal value).'),
    cfg.ListOpt('aiq_priority_order',
                default=[
                    'ops',
                    'latency',
                    'volume_count',
                    'size'
                ],
                help='Permutation of the list ["volume_count", "size", '
                     '"latency", “ops”]. Note that for volume_count and '
                     'latency, the higher the values, the less optimal the '
                     'resources. For capacity and ops, the higher the value '
                     'the more desirable the resources. If metrics are to be '
                     'considered with equal weights, concatenate the strings, '
                     'separated by ":".'
                     'An example is ["volume_count", "size", “latency:ops”] '
                     'if latency and ops want to have equal but minimum '
                     'weights, or ["volume_count:size", "latency", “ops”] '
                     'if volume_count and size have equal maximum weights. '
                     'If not provided, the default order is '
                     '["volume_count", "size", "latency", “ops”].'),
]
CONF = cfg.CONF
CONF.register_opts(active_iq_weight_opts, ACTIVE_IQ_WEIGHER_GROUP)

LOG = logging.getLogger(__name__)


class NetAppAIQWeigher(base_host.BaseHostWeigher):
    """AIQ Weigher.  Assign weights based on NetApp Active IQ tool."""

    def __init__(self, *args, **kwargs):
        super(NetAppAIQWeigher, self).__init__(*args, **kwargs)

        self.configuration = CONF[ACTIVE_IQ_WEIGHER_GROUP]

        self.host = self.configuration.aiq_hostname
        if not self.host:
            raise exception.NetappActiveIQWeigherRequiredParameter(
                config="aiq_hostname")

        self.username = self.configuration.aiq_username
        if not self.username:
            raise exception.NetappActiveIQWeigherRequiredParameter(
                config="aiq_username")

        self.password = self.configuration.aiq_password
        if not self.password:
            raise exception.NetappActiveIQWeigherRequiredParameter(
                config="aiq_password")

        self.protocol = self.configuration.aiq_transport_type
        self.port = self.configuration.aiq_port
        if not self.port:
            self.port = "80" if self.protocol == "http" else "443"

        self.ssl_verify = self.configuration.aiq_ssl_verify
        if self.ssl_verify and self.configuration.aiq_ssl_cert_path:
            self.ssl_verify = self.configuration.aiq_ssl_cert_path

        self.eval_method = self.configuration.aiq_eval_method
        self.priority_order = self.configuration.aiq_priority_order

    def _weigh_object(self, host_state, weight_properties):
        """Weight for a specific object from parent abstract class"""
        # NOTE(felipe_rodrigues): this abstract class method is not called for
        # the AIQ weigher, since it does not weigh one single object.
        raise NotImplementedError()

    def _weigh_active_iq(self, netapp_aggregates_location, weight_properties):
        """Determine host's rating based on a Active IQ."""
        size = weight_properties.get('size')
        share_type = weight_properties.get('share_type', {})
        performance_level_name = share_type.get('extra_specs', {}).get(
            'netapp:performance_service_level_name')

        # retrieves the performance service level key if a PSL name is given.
        performance_level_id = None
        if performance_level_name:
            performance_level_id = self._get_performance_level_id(
                performance_level_name)
            if not performance_level_id:
                return []

        # retrieves the equivalent active IQ keys of the pools.
        resource_keys = self._get_resource_keys(netapp_aggregates_location)
        if len(resource_keys) == 0:
            return []

        result = self._balance_aggregates(resource_keys, size,
                                          performance_level_id)

        return result

    def _get_url(self):
        """Get the base URL for REST requests."""
        host = self.host
        if ':' in host:
            host = '[%s]' % host
        return f'{self.protocol}://{host}:{self.port}/api/'

    def _get_request_method(self, method, session):
        """Returns the request method to be used in the REST call."""

        request_methods = {
            'post': session.post,
            'get': session.get,
            'put': session.put,
            'delete': session.delete,
            'patch': session.patch,
        }
        return request_methods[method]

    def _get_session_method(self, method):
        """Get the REST method from the session."""

        # NOTE(felipe_rodrigues): request resilient of temporary network
        # failures (like name resolution failure), retrying until 5 times.
        _session = requests.Session()
        max_retries = retry.Retry(total=5, connect=5, read=2, backoff_factor=1)
        adapter = HTTPAdapter(max_retries=max_retries)
        _session.mount('%s://' % self.protocol, adapter)

        _session.auth = auth.HTTPBasicAuth(self.username, self.password)
        _session.verify = self.ssl_verify
        _session.headers = {}

        return self._get_request_method(method, _session)

    def _call_active_iq(self, action_path, method, body=None):
        """Call the Active IQ REST API."""
        rest_method = self._get_session_method(method)
        url = self._get_url() + action_path

        msg_args = {
            "method": method.upper(),
            "url": url,
            "body": body,
        }
        LOG.debug("REQUEST: %(method)s %(url)s Body=%(body)s", msg_args)

        response = rest_method(url, json=body)

        code = response.status_code
        response_body = response.content
        msg_args = {
            "code": code,
            "body": response_body,
        }
        LOG.debug("RESPONSE: %(code)s Body=%(body)s", msg_args)

        return code, response_body

    def _get_performance_level_id(self, performance_level_name):
        """Gets the ID of a performance level name."""
        psl_endpoint = (f'storage-provider/performance-service-levels?'
                        f'name={performance_level_name}')
        try:
            code, res = self._call_active_iq(psl_endpoint, "get")
        except Exception as e:
            LOG.error("Could not retrieve the key of the performance service "
                      "level named as '%(psl)s'. Skipping the weigher. "
                      "Error: %(error)s",
                      {'psl': performance_level_name, 'error': e})
            LOG.error(e)
            return None

        if code != 200:
            LOG.error("Could not retrieve the key of the performance service "
                      "level named as '%(psl)s'. Skipping the weigher.",
                      {'psl': performance_level_name})
            return None

        res = jsonutils.loads(res) if res else {}
        psl_list = res.get('records', [])
        if len(psl_list) == 0:
            LOG.error("Could not found any performance service level named "
                      "as '%s'. Skipping the weigher.", performance_level_name)
            return None

        return psl_list[0].get("key", None)

    def _get_aggregate_identifier(self, aggr_name, cluster_name):
        """Returns the string identifier of an aggregate on a cluster."""
        return f'{aggr_name}:{cluster_name}'

    def _get_resource_keys(self, netapp_aggregates_location):
        """Map the aggregates names to the AIQ resource keys."""
        aggregate_endpoint = 'datacenter/storage/aggregates'

        try:
            code, res = self._call_active_iq(aggregate_endpoint, "get")
        except Exception as e:
            LOG.error("Could not retrieve the aggregates resource keys. "
                      "Skipping the weigher. Error: %s", e)
            LOG.error(e)
            return []

        if code != 200:
            LOG.error("Could not retrieve the aggregates resource keys. "
                      "Skipping the weigher.")
            return []

        res = jsonutils.loads(res) if res else {}
        aggr_map = {}
        for aggr in res.get('records', []):
            identifier = self._get_aggregate_identifier(
                aggr["name"], aggr["cluster"]["name"])
            aggr_map[identifier] = aggr["key"]

        # we must keep the lists with the same order.
        resource_keys = []
        found_pool_keys = []
        for identifier in netapp_aggregates_location:
            if identifier in aggr_map:
                found_pool_keys.append(identifier)
            # If a pool could not be found, it is marked as resource key 0.
            resource_keys.append(aggr_map.get(identifier, 0))

        LOG.debug("The following pools will be evaluated by Active IQ: %s",
                  found_pool_keys)

        return resource_keys

    def _balance_aggregates(self, resource_keys, size, performance_level_uuid):
        """Call AIQ to generate the weights of each aggregate."""
        balance_endpoint = 'storage-provider/data-placement/balance'
        body = {
            "capacity": f'{size}GB',
            "eval_method": self.eval_method,
            # NOTE(felipe_rodrigues): from Active IQ documentation, the
            # opt_method only works as 0.
            "opt_method": 0,
            "priority_order": self.priority_order,
            "separate_flag": False,
            # NOTE(felipe_rodrigues): remove the keys marked with 0, since they
            # are not found the pool keys.
            "resource_keys": [key for key in resource_keys if key != 0],
        }
        if performance_level_uuid:
            body["ssl_key"] = performance_level_uuid

        try:
            code, res = self._call_active_iq(
                balance_endpoint, "post", body=body)
        except Exception as e:
            LOG.error("Could not balance the aggregates. Skipping the "
                      "weigher. Error: %s", e)
            LOG.error(e)
            return []

        if code != 200:
            LOG.error("Could not balance the aggregates. Skipping the "
                      "weigher.")
            return []

        res = jsonutils.loads(res) if res else []
        weight_map = {}
        for aggr in res:
            weight_map[aggr["key"]] = aggr["scores"]["total_weighted_score"]

        # it must keep the lists with the same order.
        weights = []
        for key in resource_keys:
            weights.append(weight_map.get(key, 0.0))

        return weights

    def weigh_objects(self, weighed_obj_list, weight_properties):
        """Weigh multiple objects using Active IQ."""
        netapp_aggregates_location = []
        for obj in weighed_obj_list:

            # if at least one host is not from NetApp, the entire weigher is
            # skipped.
            if obj.obj.vendor_name != "NetApp":
                LOG.debug(
                    "Skipping Active IQ weigher given that some backends "
                    "are not from NetApp.")
                return []
            else:
                cluster_name = obj.obj.capabilities.get("netapp_cluster_name")
                aggr_name = obj.obj.pool_name
                netapp_aggregates_location.append(
                    self._get_aggregate_identifier(aggr_name, cluster_name))

        result = self._weigh_active_iq(
            netapp_aggregates_location, weight_properties)

        LOG.debug("Active IQ weight result: %s", result)
        return result
